@@ -331,15 +331,19 @@ void std_eparam(genericptr gen, char *key)
 
 void indicateparams(genericptr thiselem)
 {
-  int k; /* , i, j, which; (jdk) */
+   int k;
    oparamptr ops;
    eparamptr epp;
+   genericptr *pgen;
 
    if (thiselem != NULL) {
       for (epp = thiselem->passed; epp != NULL; epp = epp->next) { 
 	 ops = match_param(topobject, epp->key);
 	 if (ops == NULL) continue;	/* error condition */
-	 k = epp->pdata.pointno;
+	 if (ELEMENTTYPE(thiselem) == PATH)
+	    k = epp->pdata.pathpt[1];
+	 else
+	    k = epp->pdata.pointno;
 	 switch(ops->which) {
 	    case P_POSITION: case P_POSITION_X: case P_POSITION_Y:
 	       switch(thiselem->type) {
@@ -358,6 +362,13 @@ void indicateparams(genericptr thiselem)
 		  case SPLINE:
 	             UDrawCircle(&TOSPLINE(&thiselem)->ctrl[k], ops->which);
 		     break;
+		  case PATH:
+		     pgen = ((pathptr)thiselem)->plist + epp->pdata.pathpt[0];
+		     if (ELEMENTTYPE(*pgen) == POLYGON)
+	                 UDrawCircle(TOPOLY(pgen)->points + k, ops->which);
+		     else	/* spline */
+	                 UDrawCircle(&TOSPLINE(pgen)->ctrl[k], ops->which);
+		     break;
 	       }
 	       break;
 	 }
@@ -456,7 +467,7 @@ void setparammarks(genericptr thiselem)
 
 int opsubstitute(objectptr thisobj, objinstptr pinst)
 {
-   genericptr *eptr, thiselem;
+   genericptr *eptr, *pgen, thiselem;
    int k, ival; /* i, j, (jdk) */
    char *key;
    float fval;
@@ -550,11 +561,25 @@ int opsubstitute(objectptr thisobj, objinstptr pinst)
 	    }
 	 }
 
-	 k = epp->pdata.pointno;
+	 if (ELEMENTTYPE(thiselem) == PATH)
+	    k = epp->pdata.pathpt[1];
+	 else
+	    k = epp->pdata.pointno;
          switch(ops->which) {
 	    case P_POSITION_X:
 	       retval = max(retval, 1);
 	       switch(thiselem->type) {
+		  case PATH:
+		     pgen = TOPATH(&thiselem)->plist + epp->pdata.pathpt[0];
+		     if (ELEMENTTYPE(*pgen) == POLYGON) {
+			setpt = TOPOLY(pgen)->points + k;
+			setpt->x = ival;
+		     }
+		     else { /* spline */
+			TOSPLINE(pgen)->ctrl[k].x = ival;
+			needrecalc = True;
+		     }
+		     break;
 	          case POLYGON:
 	             setpt = TOPOLY(&thiselem)->points + k;
 		     setpt->x = ival;
@@ -577,6 +602,17 @@ int opsubstitute(objectptr thisobj, objinstptr pinst)
 	    case P_POSITION_Y:
 	       retval = max(retval, 1);
 	       switch(thiselem->type) {
+		  case PATH:
+		     pgen = TOPATH(&thiselem)->plist + epp->pdata.pathpt[0];
+		     if (ELEMENTTYPE(*pgen) == POLYGON) {
+			setpt = TOPOLY(pgen)->points + k;
+			setpt->y = ival;
+		     }
+		     else { /* spline */
+			TOSPLINE(pgen)->ctrl[k].y = ival;
+			needrecalc = True;
+		     }
+		     break;
 	          case POLYGON:
 	             setpt = TOPOLY(&thiselem)->points + k;
 		     setpt->y = ival;
@@ -714,6 +750,12 @@ int opsubstitute(objectptr thisobj, objinstptr pinst)
 	    case SPLINE:
 	       calcspline((splineptr)thiselem);
 	       break;
+	    case PATH:
+	       for (pgen = ((pathptr)thiselem)->plist; pgen < ((pathptr)thiselem)->plist
+				+ ((pathptr)thiselem)->parts; pgen++)
+		  if (ELEMENTTYPE(*pgen) == SPLINE)
+		     calcspline((splineptr)*pgen);
+	       break;
 	 }
       }
    }
@@ -784,12 +826,12 @@ Boolean has_param(genericptr celem)
 
 void pwriteback(objinstptr thisinst)
 {
-   genericptr *eptr, thiselem;
+   genericptr *eptr, *pgen, thiselem;
    objectptr thisobj;
    objinstptr pinst;
    eparamptr epp;
    oparamptr ops, ips;
-   int k, *destivalptr, found; /*  i, j, (jdk) */
+   int k, type, *destivalptr, found;
    XPoint *setpt;
    Boolean changed, need_redraw = FALSE;
    union {
@@ -823,7 +865,10 @@ void pwriteback(objinstptr thisinst)
          for (epp = thiselem->passed; epp != NULL; epp = epp->next) {
 	    if (!strcmp(epp->key, ops->key)) {
 	       found++;
-	       k = epp->pdata.pointno;
+	       if (ELEMENTTYPE(thiselem) == PATH)
+		  k = epp->pdata.pathpt[1];
+	       else
+	          k = epp->pdata.pointno;
                switch(ops->which) {
 	          case P_POSITION_X:
 	             switch(thiselem->type) {
@@ -843,6 +888,15 @@ void pwriteback(objinstptr thisinst)
 	                case SPLINE:
 		           wtemp.ival = TOSPLINE(eptr)->ctrl[k].x;
 		           break;
+			case PATH:
+			   pgen = TOPATH(eptr)->plist + epp->pdata.pathpt[0];
+			   if (ELEMENTTYPE(*pgen) == POLYGON) {
+	                      setpt = TOPOLY(pgen)->points + k;
+		              wtemp.ival = setpt->x;
+			   }
+			   else
+		              wtemp.ival = TOSPLINE(pgen)->ctrl[k].x;
+			   break;
 	             }
 	             break;
 	          case P_POSITION_Y:
@@ -863,6 +917,15 @@ void pwriteback(objinstptr thisinst)
 	                case SPLINE:
 		           wtemp.ival = TOSPLINE(eptr)->ctrl[k].y;
 		           break;
+			case PATH:
+			   pgen = TOPATH(eptr)->plist + epp->pdata.pathpt[0];
+			   if (ELEMENTTYPE(*pgen) == POLYGON) {
+	                      setpt = TOPOLY(pgen)->points + k;
+		              wtemp.ival = setpt->y;
+			   }
+			   else
+		              wtemp.ival = TOSPLINE(pgen)->ctrl[k].y;
+			   break;
 	             }
 	             break;
 	          case P_STYLE:
@@ -956,6 +1019,9 @@ void pwriteback(objinstptr thisinst)
 		     wtemp.ival = thiselem->color;
 		     break;
                }
+	       type = (ips != NULL) ? ips->type : ops->type;
+	       if (type != XC_FLOAT && type != XC_INT) break;
+
 	       destivalptr = (ips != NULL) ? &ips->parameter.ivalue
 			: &ops->parameter.ivalue;
 	       if ((!changed) && (wtemp.ival != *destivalptr)) {
@@ -1013,11 +1079,17 @@ void replaceparams(objinstptr thisinst)
 	    }
 	    break;
 	 case XC_EXPR:
-	    if (strcmp(ops->parameter.expr, ips->parameter.expr)) {
+	    /* Expression parameters should be replaced *only* if the
+	     * instance value is also an expression, and not an evaluated
+	     * result.
+	     */
+	    if ((ips->type == XC_EXPR) &&
+			strcmp(ops->parameter.expr, ips->parameter.expr)) {
 	       free(ops->parameter.expr);
 	       ops->parameter.expr = ips->parameter.expr;
 	       free_instance_param(thisinst, ips);
 	    }
+	    break;
 	 case XC_INT: case XC_FLOAT:
 	    if (ops->parameter.ivalue != ips->parameter.ivalue) {
 	       ops->parameter.ivalue = ips->parameter.ivalue;
@@ -1090,13 +1162,9 @@ void resolveparams(objinstptr thisinst)
 	    }
 	    break;
 	 case XC_EXPR:
-	    /* Always retain the last evaluated result, unless	*/
-	    /* the instance redefines the expression.		*/
-	    if (ips->type == XC_EXPR) {
-	       if (!strcmp(ops->parameter.expr, ips->parameter.expr)) {
-	          free(ips->parameter.expr);
-	          free_instance_param(thisinst, ips);
-	       }
+	    if (!strcmp(ops->parameter.expr, ips->parameter.expr)) {
+	       free(ips->parameter.expr);
+	       free_instance_param(thisinst, ips);
 	    }
 	    break;
 	 case XC_INT: case XC_FLOAT:
@@ -1127,7 +1195,7 @@ eparamptr copyeparam(eparamptr cepp, genericptr thiselem)
    if (IS_OBJINST(thiselem) && (cepp->pdata.refkey != NULL))
       newepp->pdata.refkey = strdup(cepp->pdata.refkey);
    else
-      newepp->pdata.pointno = cepp->pdata.pointno;
+      newepp->pdata.pointno = cepp->pdata.pointno;  /* also covers pathpt[] */
    return newepp;
 }
 
@@ -1197,12 +1265,15 @@ void copyparams(objinstptr destinst, objinstptr sourceinst)
       psource = sourceinst->params;
 
    for (cops = psource; cops != NULL; cops = cops->next) {
-      if (((ips = match_instance_param(destinst, cops->key)) == NULL) ||
-		(cops->type == XC_EXPR && ips->type != XC_EXPR)) {
-         newops = copyparameter(cops);
-	 newops->next = destinst->params;
-	 destinst->params = newops;
+      if ((ips = match_instance_param(destinst, cops->key)) == NULL) {
+	 if ((cops->type != XC_EXPR) || (destinst == sourceinst)) {
+            newops = copyparameter(cops);
+	    newops->next = destinst->params;
+	    destinst->params = newops;
+	 }
       }
+      else if ((cops->type == XC_EXPR) && (ips->type != XC_EXPR))
+	 free_instance_param(destinst, ips);
    }
 }
 
@@ -1212,25 +1283,40 @@ void copyparams(objinstptr destinst, objinstptr sourceinst)
 /* The expression may have either a numeric or string result.	*/
 /* the proper "which" value is passed as an argument.		*/
 /*								*/
-/* Return FALSE if unsuccessful, TRUE otherwise			*/
+/* Return NULL if unsuccessful, the parameter key (which may	*/
+/* have been modified by checkvalidname()) otherwise.		*/
 /*--------------------------------------------------------------*/
 
-Boolean makeexprparam(objectptr refobject, char *key, char *value, int which)
+char *makeexprparam(objectptr refobject, char *key, char *value, int which)
 {
    oparamptr newops;
-   char *newkey;
+   char *newkey, stkey[20];
+   int pidx;
 
    /* Check against object names, which are reserved words */
 
-   newkey = checkvalidname(key, NULL);
-   if (newkey == NULL) newkey = key;
+   if (key == NULL) {
+      strcpy(stkey, getnumericalpkey(which));
+      pidx = 0;
+      while (check_param(refobject, stkey)) {
+         pidx++;
+         sprintf(stkey, "%s%d", getnumericalpkey(which), pidx);
+      }
+      newkey = stkey;
+   }
+   else {
+      /* Check parameter key for valid name syntax */
 
-   /* Ensure that no two parameters have the same name! */
+      newkey = checkvalidname(key, NULL);
+      if (newkey == NULL) newkey = key;
 
-   if (check_param(refobject, newkey)) {
-      Wprintf("There is already a parameter named %s!", newkey);
-      if (newkey != key) free(newkey);
-      return FALSE;
+      /* Ensure that no two parameters have the same name! */
+
+      if (check_param(refobject, newkey)) {
+         Wprintf("There is already a parameter named %s!", newkey);
+         if (newkey != key) free(newkey);
+         return NULL;
+      }
    }
 
    newops = make_new_parameter(newkey);
@@ -1240,9 +1326,9 @@ Boolean makeexprparam(objectptr refobject, char *key, char *value, int which)
    newops->which = which;
    newops->parameter.expr = strdup(value);
    incr_changes(refobject);
-   if (newkey != key) free(newkey);
+   if ((newkey != key) && (newkey != stkey)) free(newkey);
 
-   return TRUE;
+   return newops->key;
 }
 
 /*------------------------------------------------------------------*/
@@ -1316,6 +1402,27 @@ Boolean makestringparam(objectptr refobject, char *key, stringpart *strptr)
    return TRUE;
 }
 
+/*--------------------------------------------------------------*/
+/* Return the built-in parameter key corresponding to a 	*/
+/* parameter type (as defined in xcircuit.h).			*/
+/*								*/
+/* Numerical parameters have designated keys to avoid the	*/
+/* necessity of having to specify a key, to avoid conflicts	*/
+/* with PostScript predefined keys, and other good reasons.	*/
+/*--------------------------------------------------------------*/
+
+char *getnumericalpkey(u_int mode)
+{
+   static char *param_keys[] = {
+	"p_gps", "p_str", "p_xps", "p_yps", "p_sty", "p_jst", "p_an1",
+	"p_an2", "p_rad", "p_axs", "p_rot", "p_scl", "p_wid", "p_col",
+	"p_bad"
+   };
+
+   if (mode < 0 || mode > 13) return param_keys[14];
+   return param_keys[mode];
+}
+
 /*--------------------------------------------------------------*/
 /* Make a numerical (integer or float) parameter.		*/
 /* If "key" is non-NULL, then the parameter key will be set	*/
@@ -1326,6 +1433,7 @@ Boolean makestringparam(objectptr refobject, char *key, stringpart *strptr)
 
 void makenumericalp(genericptr *gelem, u_int mode, char *key, short cycle)
 {
+   genericptr pgen;
    oparamptr ops, newops;
    eparamptr epp;
    XPoint *pptr;
@@ -1333,17 +1441,6 @@ void makenumericalp(genericptr *gelem, u_int mode, char *key, short cycle)
    int pidx;
    short loccycle = cycle;
 
-   /* Numerical parameters have designated keys to avoid the necessity	*/
-   /* of having to specify a key, to avoid conflicts with PostScript	*/
-   /* predefined keys, and other good reasons.				*/
-
-   /* The order of these type names must match the enumeration in xcircuit.h */
-
-   static char *param_keys[] = {
-	"p_gps", "p_str", "p_xps", "p_yps", "p_sty", "p_jst", "p_an1",
-	"p_an2", "p_rad", "p_axs", "p_rot", "p_scl", "p_wid", "p_col", NULL  /* (jdk) */
-   };
-
    /* Parameterized strings are handled by makeparam() */
 
    if (IS_LABEL(*gelem) && mode == P_SUBSTRING) {
@@ -1368,11 +1465,11 @@ void makenumericalp(genericptr *gelem, u_int mode, char *key, short cycle)
       if (keyptr == NULL) keyptr = key;
    }
    else {
-      strcpy(new_key, param_keys[mode]);
+      strcpy(new_key, getnumericalpkey(mode));
       pidx = 0;
       while (check_param(topobject, new_key)) {
          pidx++;
-         sprintf(new_key, "%s%d", param_keys[mode], pidx);
+         sprintf(new_key, "%s%d", getnumericalpkey(mode), pidx);
       }
       keyptr = new_key;
    }
@@ -1403,6 +1500,9 @@ void makenumericalp(genericptr *gelem, u_int mode, char *key, short cycle)
 			"a parameter of a different type\n");
 	 goto param_done;
       }
+      else if (ops->type == XC_EXPR)
+	 goto param_done;
+
       if ((newops = match_instance_param(areawin->topinstance, keyptr)) == NULL) {
          newops = make_new_parameter(keyptr);
          newops->next = areawin->topinstance->params;
@@ -1417,7 +1517,6 @@ void makenumericalp(genericptr *gelem, u_int mode, char *key, short cycle)
 	 /* should redraw.						*/
          drawarea(NULL, NULL, NULL);
          newops = NULL;
-	 goto param_done;
       }
    }
 
@@ -1500,12 +1599,12 @@ void makenumericalp(genericptr *gelem, u_int mode, char *key, short cycle)
 	       case P_POSITION_X:
 	          pptr = TOPOLY(gelem)->points + loccycle;
 	          newops->parameter.ivalue = (int)pptr->x;
-	          epp->pdata.pointno = TOPOLY(gelem)->cycle;
+	          epp->pdata.pointno = loccycle;
 	          break;
 	       case P_POSITION_Y:
 	          pptr = TOPOLY(gelem)->points + loccycle;
 	          newops->parameter.ivalue = (int)pptr->y;
-	          epp->pdata.pointno = TOPOLY(gelem)->cycle;
+	          epp->pdata.pointno = loccycle;
 	          break;
 	       case P_STYLE:
 	          newops->parameter.ivalue = (int)TOPOLY(gelem)->style;
@@ -1522,12 +1621,12 @@ void makenumericalp(genericptr *gelem, u_int mode, char *key, short cycle)
 	       case P_POSITION_X:
 	          pptr = TOSPLINE(gelem)->ctrl + loccycle;
 	          newops->parameter.ivalue = (int)pptr->x;
-	          epp->pdata.pointno = TOSPLINE(gelem)->cycle;
+	          epp->pdata.pointno = loccycle;
 	          break;
 	       case P_POSITION_Y:
 	          pptr = TOSPLINE(gelem)->ctrl + loccycle;
 	          newops->parameter.ivalue = (int)pptr->y;
-	          epp->pdata.pointno = TOSPLINE(gelem)->cycle;
+	          epp->pdata.pointno = loccycle;
 	          break;
 	       case P_STYLE:
 	          newops->parameter.ivalue = (int)TOSPLINE(gelem)->style;
@@ -1539,6 +1638,19 @@ void makenumericalp(genericptr *gelem, u_int mode, char *key, short cycle)
 	    }
 	    break;
          case PATH:
+	    if (loccycle == -1 && (mode == P_POSITION_X || mode == P_POSITION_Y)) {
+	       pgen = getsubpart(TOPATH(gelem));
+	       pidx = getsubpartindex(TOPATH(gelem));
+	       if (ELEMENTTYPE(pgen) == POLYGON)
+	          loccycle = ((polyptr)pgen)->cycle;
+	       else
+	          loccycle = ((splineptr)pgen)->cycle;
+	    }
+	    else {
+	       Fprintf(stderr, "Can't parameterize a path point from "
+			"the command line.\n");
+	       break;
+	    }
 	    switch(mode) {
 	       case P_STYLE:
 	          newops->parameter.ivalue = (int)TOPATH(gelem)->style;
@@ -1547,6 +1659,24 @@ void makenumericalp(genericptr *gelem, u_int mode, char *key, short cycle)
 	          newops->type = XC_FLOAT;
 	          newops->parameter.fvalue = TOPATH(gelem)->width;
 	          break;
+	       case P_POSITION_X:
+		  newops->type = XC_INT;
+		  if (ELEMENTTYPE(pgen) == POLYGON)
+		     newops->parameter.ivalue = ((polyptr)pgen)->points[loccycle].x;
+		  else
+		     newops->parameter.ivalue = ((splineptr)pgen)->ctrl[loccycle].x;
+	          epp->pdata.pathpt[1] = loccycle;
+	          epp->pdata.pathpt[0] = pidx;
+		  break;
+	       case P_POSITION_Y:
+		  newops->type = XC_INT;
+		  if (ELEMENTTYPE(pgen) == POLYGON)
+		     newops->parameter.ivalue = ((polyptr)pgen)->points[loccycle].y;
+		  else
+		     newops->parameter.ivalue = ((splineptr)pgen)->ctrl[loccycle].y;
+	          epp->pdata.pathpt[1] = loccycle;
+	          epp->pdata.pathpt[0] = pidx;
+		  break;
 	    }
 	    break;
       }
@@ -1709,7 +1839,7 @@ void makeparam(labelptr thislabel, char *key)
       return;
    }
 
-   /* check parameter against reserved PostScript names */
+   /* Check parameter for valid name syntax */
 
    newkey = checkvalidname(key, NULL);
    if (newkey == NULL) newkey = key;
